/*
 *  Maintainer : Dries Pruimboom  (dries[at]irssystems.nl)
 *  
 *  Disclaimer : I can not be blamed for anything that goes wrong caused by the 
 *               use of this program, use at your own risk !!
 *               
 *  Copyrights : This is improve-ware.
 *               Improve-ware licence : you can copy change etc. this program
 *               you MUST however mail any improvements you make to this program
 *               to the maintainer of the program.
 *               
 *   
 */
 
#include <stdio.h>
#include <stdlib.h>
#include "sys_nodelib.h"

#define VERSION_STRING "V0.0"

#define FALSE 0
#define TRUE  1

/*
 *  Globals that contain settings etc :
 */
char StartSymbol[80];
char InputFile[200];
/*
 *  End of globals that contain settings etc
 */

/*
 *  Typedefs and list object routines
 */
 
typedef struct StringNode{
                            Node __internal;
			    char *str;
                         } StringNode;
			 
/*
 *  Keeps all called routines
 */
NodeList CallList;


StringNode *NewStringNode(char *str)
{
    StringNode *hlp;
    
    hlp = malloc(sizeof(StringNode));
      if (hlp)
       {
           hlp->str = malloc(strlen(str)+2);
	   if (hlp)
	   {
	      strcpy(hlp->str,str);
	   }
	   else
	   {
	      free(hlp);
	      hlp=NULL;
	   }
       }
    return(hlp);
}
 
void DeleteStringNode(void *node)
{
   StringNode *hlp;
   
   hlp = (StringNode *) node;
   if (hlp)
   {
      if (hlp->str)
      {
          free(hlp->str);
         
      }
      free(hlp);
   }
     
}



/*
 *  Find a string in a string list
 */
int FindString(NodeList *l,char *str)
{
    StringNode *hlp;
    int res=0;
    
    hlp = (StringNode *)FirstNode(l);
    
    while (hlp && (res==0))
    {
       if (strcmp(hlp->str,str)==0)
       {
          res=1;
       }
       else
       {
          hlp = (StringNode *)NextNode((Node*)hlp);
       }
    }
    
    return(res);
}

/*
 *  End of list object routines
 */

/*
 *  Read a word from a string.
 */
int GetWord(char *str,char *res,int maxlen)
{
   int i,x=0;
   i=0;
   
   maxlen--;
   while (isspace(str[i]) && str[i])
   {
          i++;
   }

   while (!isspace(str[i]) && (str[i]) && (x<maxlen))
   {
          res[x]=str[i];
          x++;
          i++;
   }
   res[x]='\0';
   return(res[0]);
}


/*
 *  Find a file line containing "DEF : searchname"
 */
int FindDef(FILE *f,char *search)
{
   char line[400];
   char word[100];
   int done=FALSE;
   char *ptr;
   
   fseek(f,0,SEEK_SET);
   
   while (!done && (fgets(line,400,f)))
   {
      ptr = strstr(line,"DEF :");
      if (ptr)
      {
         GetWord(ptr+5,word,100);
         if (strcmp(word,search)==0)
	 {
	    done=TRUE;
         }
      }
   }
   
   return(done);
}

int GetUse(FILE *f,char *name)
{
   char line[120];
   int res=0;
   char *ptr;
   
   if (fgets(line,120,f))
   {
      ptr = strstr(line,"USE :");
      if (ptr)
      {
         GetWord(ptr+5,name,80);
         res=1;
      }
   }
   
   return(res);
}
/*
 *  Tab a few spaces on a line
 */
void Tab(int indentlevel)
{
   int x=0;
   
   for (x=0;x<indentlevel-1;x++)
   { 
      printf("| ");
   }

   if (indentlevel>0)
   {
      printf("+-");
   }
}

/*
 *  This is what does it ! 
 */
int IndentLevel=0;
void MakeCallTree(FILE *f,char *symbol)
{
   unsigned long currentpos;
   char routine[80];
   int done=0;
   StringNode *sn;
   
   /*
    *  Place tab level
    */
   Tab(IndentLevel);
   /*
    *  Print current symbol
    */
   printf("%s\n",symbol);
   
   /*
    *  Add current symbol to the called list to detect recursion
    */
   sn = NewStringNode(symbol);
   AddBackNode(&CallList,(Node *)sn);
   
 
   /*
    *  Try to find definition
    */
   if (FindDef(f,symbol))
   {
      /*
       *  Find all uses inside this function
       */
      while (!done)
      {
         if (GetUse(f,routine))
	 {
	    /*
	     *  For each used routine make a call tree
	     */
	    currentpos = ftell(f);
	    IndentLevel++;
	    if (!FindString(&CallList,routine))
	    {
	       MakeCallTree(f,routine);
	    }
	    else
	    {
	        /*
		 *   give recursion message
		 */
                Tab(IndentLevel);
                printf("%s :: RECURSING\n",symbol);
	    }
	    IndentLevel--;
	    fseek(f,currentpos,SEEK_SET);
	 }
	 else
	 {
	    done=1;
	 }
      }  
   }
   else
   {
      if (IndentLevel==0)
      {
         printf("Could not find start symbol %s\n",symbol);
      }
   }
   
   /*
    *  Remove symbol from call_list
    */
   sn = (StringNode*)DeleteLast(&CallList);
}
 

int main(int argc,char **argv)
{
   int x;
   FILE *f;
   
   /*
    *  Version string
    */
   printf("Version %s\n",VERSION_STRING);
   
   /*
    *   First Init   
    */
   strcpy(StartSymbol,"main");
   InitNodeList(&CallList);
   
   /*
    *   End of first init  
    */
    
   x=1;
   while (x<argc)
   {
       if (strcasecmp(argv[x],"-start")==0)
       {
          x++;
	  strcpy(StartSymbol,argv[x]);
	  printf("Start Symbol set to : %s\n",StartSymbol);
       }
       else
       {
           strcpy(InputFile,argv[x]);
           printf("file : %s\n",InputFile);          
       }
       x++;
   }


   f = fopen(InputFile,"r");
   if (f)
   {
      MakeCallTree(f,StartSymbol);
   }
   else
   {
      printf("Could not open input file %s\n",InputFile);
   }

   exit(0);
}

