/*///////////////*/
/*//////////////  NodeLib , contains routines for linked lists.*/
/*/////////////             embed the types NodeList and Node in your own*/
/*////////////              structs to make easy use of adding and removing*/
/*///////////               items in the list. (this is more or less stolen*/
/*//////////                from the AMIGA operating system.)*/
/*/////////*/
/*////////*/
/*///////*/
/*//////*/
/*/////*/
/*////*/
/*///*/
/*//*/
/*/*/
/**/
/**/
/*  EXAMPLE :*/
/**/
/*..........................................................................*/
/**/
/*       typedef struct MyList { NodeList _liststruct;*/
/*                               int x;*/
/*                               int y;*/
/*                               int hlp;*/
/*                             } MyList;*/
/**/
/*       typedef struct MyItem { Node _node;*/
/*                               int width;*/
/*                               int height;*/
/*                               int color;*/
/*                             } MyItem;*/
/**/
/**/
/*       main(void)*/
/*       {*/
/*           MyList list;*/
/*           MyItem *item;*/
/**/
/*           InitNodeList(&list);*/
/**/
/*           item = MyAlloc(sizeof(MyItem));*/
/**/
/*           AddBackNode(&list,item);            // add item as last in list*/
/**/
/*           item = MyAlloc(sizeof(MyItem));*/
/**/
/*           AddBackNode(&list,item);            // add item as last in list*/
/**/
/*           item = MyAlloc(sizeof(MyItem));*/
/**/
/*           InsertNode(&list,item,list.first);  // add item after the first item*/
/**/
/*           DeleteNodeList(&list,MyFree);      // delete all nodes using the free command*/
/**/
/*       }*/
/**/
/*..........................................................................*/
/**/
/**/
/**/
#include <stdio.h>       /* Needed for NULL pointer*/
#include <stdlib.h>
#include <malloc.h>


#include "sys_nodelib.h"


/**/
/* Initialize the nodelist to a NULL list*/
/**/
void InitNodeList(NodeList *l)
{
   l->first = NULL;
   l->last  = NULL;
   l->count = 0;    /* zero items in list*/
}

/* This routine initializes one node */
void InitNode(Node *n)
{
    n->next = NULL;
    n->prev = NULL;
}

/**/
/* Add a node at the back of the list*/
/**/
void AddBackNode(NodeList *list,Node *n)
{
   if (n == NULL)
   {
          return;
   }

   if (list->last == NULL)    /* There are no nodes in the list at all ! */
   {
          list->first = n;
          list->last  = n;
          list->count++;
          n->prev = NULL;
          n->next = NULL;
   }
   else
   {
         n->prev = list->last;  /* there are nodes, bend the pointers */
         n->next = NULL;

         list->count++;
         list->last->next = n;
         list->last = n;
   }
}

/**/
/* Add node at the front of the list*/
/**/
void AddFrontNode(NodeList *list,Node *n)
{
   if (n == NULL)
   {
          return;
   }

   /**/
   /* First node to insert (no node in this list ??)*/
   /**/
   if (list->first == NULL)
   {
          list->first = n;
          list->last  = n;
          list->count++;
          n->prev = NULL;
          n->next = NULL;
   }
   else
   {
         /**/
         /* Make new front node*/
         /**/
         n->next = list->first;
         n->prev = NULL;

         list->first->prev = n;
         list->first = n;
         list->count++;
   }
}


/**/
/* DetachNode : cut a node out of the list (node is not freed !!)*/
/**/
Node *DetachNode(NodeList *l,Node *n)
{
   /**/
   /* delete a valid node ??*/
   /**/
   if (n == NULL)
   {
          return(NULL);
   }

   /**/
   /* is this the last node ?*/
   /**/
   if (n->next != NULL)
   {
          /**/
          /* this is the last node, is it also the first (only) node ??*/
          /**/
          if (n->prev != NULL)
          {
                 n->next->prev = n->prev;
                 n->prev->next = n->next;
          }
          else
          {
                 l->first = n->next;
                 n->next->prev = NULL;
          }
   }
   else
   {
          /**/
          /* first node ??*/
          /**/
          if (n->prev != NULL)
          {
                 l->last = n->prev;
                 n->prev->next = NULL;
          }
          else
          {
                 l->first = NULL;
                 l->last  = NULL;
          }
   }

   /**/
   /* Node detached from list, decrement count*/
   /* */
   l->count--;
   /**/
   /* Node is not in a list anymore, null the next and prev pointers*/
   /**/
   n->next = NULL;
   n->prev = NULL;

   return(n);
}


/**/
/* Insert node n1 after node n2*/
/**/
void InsertNode(NodeList *l,Node *n1,Node *n2)
{
   /**/
   /* we passed a valid node ??*/
   /*  */
   if (n1 == NULL)
   {
          /**/
          /* Nothing to insert*/
          /**/
          return;
   }

   /**/
   /* if the node we want to place the node after is not there add it as front*/
   /**/
   if (n2 == NULL)
   {
          AddFrontNode(l,n1);
   }
   else
   {
          /**/
          /* add it as last node ?*/
          /**/
          if (n2->next == NULL)
          {
                 AddBackNode(l,n1);
          }
          else
          {
              /**/
              /* Add inbetween these*/
              /**/
              n1->next = n2->next;
              n1->next->prev = n1;

              n1->prev = n2;
              n2->next = n1;
              l->count++;
          }

   }
}


/**/
/* Delete a complete node list, memory is freed !.*/
/**/
/* the DeleteNode variable points to a function that handles complete*/
/*  deletion of a node.*/
/**/
void DeleteNodeList(NodeList *l,NodeFreeFunc DeleteNode)
{
   Node *n,*n1;

   if (DeleteNode==NULL)
   {
          DeleteNode = free;
   }

   n = l->first;
   while (n != NULL)
   {
          n1=n->next;

          /**/
          /* Delete the Node using a user supplied function.*/
          /**/
          DeleteNode(n);
          n=n1;
   }
   l->first = NULL;
   l->last  = NULL;
   l->count = 0;
}

/**/
/* Go from this node to the next*/
/**/
Node *NextNode(Node *i)
{
   if (i == NULL)
   {
          return(NULL);
   }

   return(i->next);
}

/**/
/* Go from this node to the previous*/
/**/
Node *PrevNode(Node *i)
{
   if (i == NULL)
   {
          return(NULL);
   }
   return(i->prev);
}

/**/
/* Delete the first node in the list (do not free the memory).*/
/**/
Node *DeleteFirst(NodeList *l)
{
   Node *res;

   res = l->first;
   if (res != NULL)
   {
          l->first = res->next;

          if (l->first != NULL)
          {
                 l->first->prev = NULL;
          }
          else
          {
                 /**/
                 /* List Must be empty*/
                 /**/
                 l->last = NULL;
          }
          l->count--;
   }
   return(res);
}



/**/
/* Delete the last node in the list (do not free the memory).*/
/**/
Node *DeleteLast(NodeList *l)
{
   Node *res;

   res = l->last;

   /**/
   /* Ther is a last node ?*/
   /**/
   if (res != NULL)
   {
          l->last = res->prev;

          /**/
          /* Last node deleted ??*/
          /**/
          if (l->last != NULL)
          {
                 l->last->next = NULL;
          }
          else
          {
                 l->first = NULL;
          }
          l->count--;
   }
   return(res);
}

/**/
/* Get pointer to first node*/
/**/
Node *FirstNode(NodeList *l)
{
   return(l->first);
}

/**/
/* Get pointer to last node*/
/**/
Node *LastNode(NodeList *l)
{
   return(l->last);
}


/**/
/* Insert a node at a sorted place in the list (if the compare function*/
/* == NULL then a insert front is done*/
/**/
/**/
/* the compare function is of the type  "int func(Node *n1,Node *n2)"*/
/* the function should return -1 if n1<n2*/
/*                             0 if n1=n2*/
/*                             1 if n1>n2*/
/* like string compare !!*/
/**/

void InsertSortedNode(NodeList *l,Node *n1,NodeCompareFunc CompareNode)
{
  int ready=0;
  Node *cur,*hlp;

        if (CompareNode == NULL)
        {
           AddFrontNode(l,n1);
        }
        else
        {
           cur = FirstNode(l);
           while ((cur != NULL) && !ready)
           {
                  /**/
                  /* If cur > n1 then insert before n1*/
                  /**/
                  if (CompareNode(cur,n1) > 0)
                  {
                         hlp = PrevNode(cur);
                         if (hlp == NULL)
                         {
                             AddFrontNode(l,n1);
                         }
                         else
                         {
                             InsertNode(l,n1,hlp);
                         }
                         ready=1;
                  }
                  else
                  {
                         cur = NextNode(cur);
                  }
           }
           
           if (!ready)
           {
                  AddBackNode(l,n1);
           }
        }
}

/**/
/* NodeCount(NodeList *l)  >> returns the number of nodes in a NodeList*/
/**/
int NodeCount(NodeList *l)
{
    return(l->count);
}
