%x comment
%x quote
%x assignment

/*
 *   Maintainer : Dries Pruimboom  (dries[at]irssystems.nl)
 *  
 *  Disclaimer : I can not be blamed for anything that goes wrong caused by the 
 *               use of this program, use at your own risk !!
 *               
 *  Copyrights : This is improve-ware.
 *               Improve-ware licence : you can copy change etc. this program
 *               you MUST however mail any improvements you make to this program
 *               to the maintainer of the program.
 *               
 *   
 */

%{
  #include "sys_nodelib.h"
  
  #define DOUBLE_DEFINES 1
  #define UNDEFINED      2
  #define CALL_LIST      4
  #define UNUSED         8
  #define USAGE          16
  #define OTHERFILE      32
  
  int DefinedRoutines=0;
  int blocknest=0;
  int OPT;
  void HandleRoutine(char *str,int use);
  void CheckRoutine(char *str);
  void MightBeUsage(char *str);

%}


DIGIT       [0-9]
LETTER      [a-zA-Z]
VALID_ID    [a-zA-Z0-9_]
ANYCHAR     [:print:]
ROUTINE     {LETTER}{VALID_ID}+[:space:]*\( 
IDENTIFIER  {LETTER}{VALID_ID}*
%%

#{IDENTIFIER}  /**/

               /* Character consts */
"'"."'"        /* no action */
"'"\\."'"      /* no action */

               /* C-style comments  */
"/*"               BEGIN(comment); 
<comment>[^*]*      /*eat all but * */  
<comment>"*"+[^/]   /* eat all * followed by a non / */
<comment>"*"+"/"   BEGIN(INITIAL);

"//".*\n         /* dit is commentaar van het nieuwe C++ type */

                  /* QUOTE */
\"                BEGIN(quote);
<quote>[^\"]       /* eat all but quotes */
<quote>"\\".      /* eat escape sequences (\n \t \" etc */
<quote>\"         BEGIN(INITIAL);

          /*
           *   See if this is use of the ID
           */
{IDENTIFIER}  if (blocknest) MightBeUsage(yytext);

"{"        blocknest++;
"}"        blocknest--;

{ROUTINE}         HandleRoutine(yytext,blocknest);

                  /* routine followed by a ');' is a proto if we are not */
                  /* in a functional block, we skip protos since there are no */
                  /* routines used in a prototype and it might really be define*/
                  /* somewhere else */
{ROUTINE}.*");"   if (blocknest) { HandleRoutine(yytext,blocknest);};



.      /**/
\n     /**/; 
%%

/***************************************************************************/
char currentfile[200];   /* the name of the file we are currently working on */

/*
 *   this structure is used to store all Routines found
 */
typedef struct Entry { 
                        Node INTERNAL;
                        char *name;
                        int usage;
                        int possibleusage;
                        int defined;
                        int system;
                        char *file;
                     } Entry;

void DeleteEntry(void *e)
{
   Entry *en;
   
   en = (Entry *)e;
   if (en)
   {
       if (en->name)
       {
          free(en->name);
       }
       if (en->file)
       {
          free(en->file);
       }
       free(en);
   }
}

int NrOfSymbols=0;

Entry *NewEntry(char *name,char *file)
{
    Entry *en;
    
    en = malloc(sizeof(Entry));
    if (en)
    {
        NrOfSymbols++;
        
        InitNode((Node *)en);
        en->usage=0;
        en->possibleusage=0;
        en->defined=0;
        en->system=0;
        en->name = malloc(strlen(name)+2);
        if (en->name)
        {
            strcpy(en->name,name);
            en->file = malloc(strlen(file)+2);
            if (en->file)
            {
               strcpy(en->file,file);
            }
            else
            { 
               DeleteEntry(en);
               en=NULL;
            }
        }
        else
        {
           DeleteEntry(en);
           en=NULL;
        }
    }
    
    return(en);
}

/*
 *   Modify the filename
 */
void SetFileName(Entry *e,char *str)
{ 
    if (e)
    {
        if (e->file)
        {
           free(e->file);
        }
        e->file = malloc(strlen(str)+2);
        if (e->file)
        {
            strcpy(e->file,str);
        }
    }
}


/*
 *   For speedy lookup, hash table
 */                   
#define HASH_SIZE 32  
#define HASH_VAL(str)  ((str[0]+str[1]) % HASH_SIZE)

NodeList HashTable[HASH_SIZE];

/*
 *   Initialize the hash table before usage
 */
void InitHashTable(void)
{
   int x=0;
   for (x=0;x<HASH_SIZE;x++)
   {
      InitNodeList(&HashTable[x]);
   }
}
                          
/*
 *   Delete all entries from the hash table
 */
void DeleteHashTable(void)
{
    int x;
    
    for(x=0;x<HASH_SIZE;x++)
    {
        DeleteNodeList(&HashTable[x],DeleteEntry);
    }
}                          
                           
/*
 *   Strip whitespace and ( from routine name.
 */ 
void GetRoutineName(char *res,char *src)
{
   int i=0;
   
   while (src[i] && (src[i]!='(') && !isspace(src[i]))
   {
       res[i]=src[i];
       i++;
   }
   res[i]='\0';
}

/*
 *   Find routine name in the hash table
 */
Entry *FindRoutine(char *name)
{
    int i;
    Entry *e;
    int done=0;
    
    i = HASH_VAL(name);
    
    e = (Entry *)FirstNode(&HashTable[i]);
    
    while (e && (!done))
    {
       if (strcmp(e->name,name)==0)
       {
          done=1;
       }
       else
       {
          e = (Entry *)NextNode((Node *)e);
       }
    }
    
    return(e);
}

void FindNotUsed(void)
{
    int i;
    Entry *e;
 
 
    if (OPT & UNUSED)
    {
       printf("\nNot used routines : \n");
       for (i=0;i<HASH_SIZE;i++)
       {
          e = (Entry *)FirstNode(&HashTable[i]);
     
          while (e)
          { 
              /*
               *   List unused routines but skip the system routines 
               */
              if (e->usage==0 && (!e->system) && (e->defined))
              {
                 if (e->possibleusage)
                 {  
                    printf("?");
                 }
                 else
                 {
                    printf(" ");
                 }
                 printf("%-34s : file %s\n",e->name,e->file);
              } 
              e = (Entry *)NextNode((Node *)e);
          }
       }
   }   


   if (OPT & UNDEFINED)
   {
       printf("\nUndefined routines : \n");
       for (i=0;i<HASH_SIZE;i++)
       {
          e = (Entry *)FirstNode(&HashTable[i]);
     
          while (e)
          { 
              if ((e->defined==0) && (!e->system))
              {
                 printf("%-34s : file %s\n",e->name,e->file);
              } 
              e = (Entry *)NextNode((Node *)e);
          }
       }
   }   

   if (OPT & USAGE)
   {
       printf("\nRoutineUsage : \n");
       for (i=0;i<HASH_SIZE;i++)
       {
          e = (Entry *)FirstNode(&HashTable[i]);
     
          while (e)
          { 
              if (!e->system && e->defined)
              {
                 printf("%4d %-34s : file %s\n",e->usage,e->name,e->file);
              }
              e = (Entry *)NextNode((Node *)e);
          }
       }
   }   

}

/*
 *   CheckRoutine(str) >> check if it is a routine that is assigned
 */
void CheckRoutine(char *str)
{
   char RoutineName[120];
   Entry *e;
   int i;
   
   GetRoutineName(RoutineName,str);
   e = FindRoutine(RoutineName);
   if (e)
   {
      e->usage++;
   }
}
/*
 *   This might be usage of a function, but it might not be also
 */
void MightBeUsage(char *str)
{
   char RoutineName[120];
   Entry *e;
   int i;
   
   GetRoutineName(RoutineName,str);

   e = FindRoutine(RoutineName);
   if (e)
   { 
      if (e->system == 0)
      {
        if (((OPT & OTHERFILE) == 0) || (strcmp(currentfile,e->file)!=0))
        { 
           e->possibleusage++;
        }
        if (OPT & CALL_LIST)
        {
           printf ("? USE : %-24s     %4d  %s\n",e->name,e->usage,currentfile);
        }
      }
   }
   else
   {
       e = NewEntry(RoutineName,currentfile);
       if (e)
       {
          e->possibleusage=1;
       }
   }
}

/*
 *   This routine is called from the scanner.
 */
void HandleRoutine(char *str,int use)
{
   char RoutineName[120];
   Entry *e;
   int i;
   
   GetRoutineName(RoutineName,str);
   if (use)
   { 
       e = FindRoutine(RoutineName);
       if (e)
       { 
          if (((OPT & OTHERFILE) == 0) || (strcmp(currentfile,e->file)!=0))
          { 
             e->usage++;
          }
       }
       else
       {
          e = NewEntry(RoutineName,currentfile);
          if (e)
          {
             
             if (((OPT & OTHERFILE) == 0) || (strcmp(currentfile,e->file)!=0))
             { 
                e->usage=1;
             }
             i=HASH_VAL(RoutineName);
             AddBackNode(&HashTable[i],(Node *)e);
          }
       }
       
       if (OPT & CALL_LIST)
       {
          if (e->system == 0)
          {
             printf ("  USE : %-24s     %4d  %s\n",e->name,e->usage,currentfile);
          }
       }
   }
   else
   {
       e = FindRoutine(RoutineName);
       if (e)
       { 
          if (e->defined)
          {
              if (OPT & DOUBLE_DEFINES)
              {
                  printf("%s already defined in %s   current file : %s\n",e->name,e->file,currentfile);
              } 
          }
          SetFileName(e,currentfile);
          e->defined++;
       }
       else
       {
          e = NewEntry(RoutineName,currentfile);
          if (e)
          {
             e->defined=1;
             i=HASH_VAL(RoutineName);
             AddBackNode(&HashTable[i],(Node *)e);
          }
       }

       if (OPT & CALL_LIST)
       {
          printf("DEF : %-24s             %s\n",e->name,e->file);
       }
       DefinedRoutines++;
   }   
}

void LoadSystemRoutines(void)
{
   FILE *f;
   char hlp[200];
   int i;
   Entry *e;
   
   f = fopen("/etc/usage.lst","r");
   if (f)
   {
      while (!feof(f))
      {
         fscanf(f,"%s",hlp);
         if (hlp[0])
         {
             i = HASH_VAL(hlp);        
             e = NewEntry(hlp,"SYSTEM");
             if (e)
             {
                e->system=1;
                AddBackNode(&HashTable[i],(Node*)e);
             } 
         }
      }
      fclose(f);
   }
   else
   {
      printf("/etc/usage.lst could not be opened, this list contains routines\n");
      printf(" that we don't wanna see in the output lists.\n\n");
   }
   NrOfSymbols=0;
}
void InitProgram(void)
{
   strcpy(currentfile,"stdin");
   OPT = UNUSED;
}

int main(int argc,char **argv)
{
   int i=1,x;
   FILE *f;
   int filesdone=0;
         
     InitProgram();
     InitHashTable();
     LoadSystemRoutines();

     printf(" (option -h for help !!)\n");

     while (i<argc)
     {
        if (argv[i][0]=='-')
        {
           for (x=1;x<strlen(argv[i]);x++)
           {
              switch(argv[i][x])
              {
                 case 'c': OPT ^= CALL_LIST; break;  
                 case 'd': OPT ^= UNDEFINED; break;
                 case 'u': OPT ^= UNUSED;break;
                 case 'n': OPT ^= USAGE;break;
                 case 'o': OPT ^= OTHERFILE;break;
                 
                 case 'f': /* read filenams from file */ 
                    i++;
                    f=fopen(argv[i],"r");
                    if (f)
                    {
                       while (!feof(f))
                       {
                          fscanf(f,"%s",currentfile);
                          yyin = fopen(currentfile,"r");
                          if (yyin)
                          {
                             blocknest=0;
                             filesdone++;
                             yylex();
                             fclose(yyin);
                          }
                       }
                       fclose(f);
                    }
                    else
                    {
                       printf("Could not open %s\n",argv[i]);
                    }
                    x=100230;
                 break;
                 default:
                   printf("Unknown commandline param : -%c\n",argv[i][x]);
                   printf("usage : usage [options] filename filename2 filename . . .\n");
                   printf("options : -c : display call list\n");
                   printf("          -d : show undifine list\n");
                   printf("          -u : don't show unused list\n");
                   printf("          -n : show use count list\n");
                   printf("          -o : mark only as used if used in another file\n");
                   printf("          -f filename : read sourcefiles from this file\n");
                   printf("\n\n");
                   printf("An attempt is made to open the file '/etc/usage.lst'\n");
                   printf("this file contains a list of routines that we don't want\n");
                   printf("in our lists. A typical file could contain lines like this :\n");
                   printf("  printf\n");
                   printf("  fprintf\n");
                   printf("  sprintf\n");
                   printf("  scanf\n");
                   printf("  fscanf\n");
                   printf("\n etc.\n");
                   printf("\nWhen using the -f filename option the file should contain\n");
                   printf("filenames of source files example : \n");
                   printf("       /home/dries/c/test.c\n");
                   printf("       /home/dries/c/lib.c\n");
                   printf("       /home/dries/c/modules.c\n");
                   printf("       /home/dries/c/other.c\n");
                   printf(" etc.\n");
              }
           }
        }
        else
        {
            blocknest=0;
            yyin = fopen(argv[i],"r");
            if (yyin)
            {
               filesdone++;
               strcpy(currentfile,argv[i]);
               yylex();
               fclose(yyin);
            }
            else
            {
                printf("Could not open %s\n",argv[i]);
            }
        }
        i++;
     }
     
     if (filesdone>0)
     {
        if (OPT & (UNUSED + UNDEFINED))
        {
           FindNotUsed();
        }
        printf("\n# Defined Routines : %d  total nr of symbols : %d\n",DefinedRoutines,NrOfSymbols);
     }
     
     DeleteHashTable();
}




   